/*
 * Copyright (C) 2021 B.R.S. Roso.
 * 
 * Xselection is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Xselection.  If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * Functions to manipulate locked memory.
 * For correct behaviour, this should be the only code in a process which uses
 * mlock and munlock on the heap. See why below.
 */

#include <stdlib.h>
#include <sys/mman.h>
#include <unistd.h>

#include "shred.h"
#include "lmem.h"

/* Allocate and lock size bytes. The address of the allocated memory is aligned
 * to the page size; this is vital for everything in this file to work. If one
 * were to mlock two bytes in a page (perhaps in separate areas of the code
 * which not be obviously related, eg if malloc squeezes two allocations into
 * the same page), then a single call to unlock, for either of the bytes,
 * unlocks the whole page. This is a problem. But if all the code which calls
 * mlock and munlock be the one contained in this file, we avoid that issue by
 * always making allocations that begin at a multiple of the page size. */
int lmalloc(struct lmem *ret, size_t size)
{
	void *mem;
	if (posix_memalign(&mem, sysconf(_SC_PAGESIZE), size))
		return 1;
	if (mlock(mem, size)) {
		free(mem);
		return 1;
	}
	ret->size = size;
	ret->mem = mem;

	return 0;
}

/* Overwrite mem with zeros, unlock it and free it. */
int ldestroy(struct lmem *mem)
{
	void * const p = mem->mem;
	const size_t size = mem->size;

	if (p == NULL)
		return 0;

	shred(p, size);
	if (munlock(p, size))
		return 1;
	free(p);

	return 0;
}

/* Resize mem to new_size. If new_size be smaller than the size of mem, the data
 * in mem will be truncated. Note that, here, we always end up allocating
 * new_size bytes and copying the data from mem to the new area, then setting
 * mem to refer to the new area. This is the case even when we want to reduce
 * the size of mem, due to the fact that we cannot use realloc without risking
 * it leaking the data to unlocked memory. */
int lrealloc(struct lmem *mem, size_t new_size)
{
	const size_t old_size = mem->size;
	struct lmem mem_new;

	if (lmalloc(&mem_new, new_size))
		return 1;
	char * const dst = mem_new.mem;
	char * const src = mem->mem;
	const size_t n = new_size > old_size ? old_size : new_size;
	size_t i;
	for (i = 0; i < n; i++)
		dst[i] = src[i];
	if (ldestroy(mem))
		return 2;
	*mem = mem_new;

	return 0;
}

/* Allocates the minimal amount of memory which can be locked. */
int lnew(struct lmem *ret)
{
	return lmalloc(ret, sysconf(_SC_PAGESIZE));
}

/* Grows mem by the minimal amount of memory which can be locked. */
int lgrow(struct lmem *mem)
{
	return lrealloc(mem, mem->size + sysconf(_SC_PAGESIZE));
}
