/*
 * Copyright (C) 2021 B.R.S. Roso.
 * 
 * Xselection is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Xselection.  If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * This small library implements a simple signal handling mechanism. The
 * intended usage is for programs which want to prevent certain chosen signals
 * that would cause process termination from doing so immediately; instead,
 * they are caught and the standard action of the first (and only the first)
 * signal received can be executed when convenient for the program; however,
 * this differs from blocking the signals in that system calls like read, write
 * and open which may return upon reception of a signal will still do so. This
 * may be used to have the exit code of the program reflect the fact that the
 * program exited due to the signal it received. See rdseln or lbuf for example
 * usage.
 *
 * To mark signals to be caught, use sig_catch or sig_catch_v; then, to detect
 * if a signal has been caught, test the global variable sig_caught; then, when
 * the default action of the signal be desired, use sig_caught_raise. To know
 * which signal has been caught, use sig_caught_num_get.
 *
 * The limitation of the code here is that, once any of the signals registered
 * to be caught is received, then none of the other signals has any effect for
 * the rest of the lifetime of the program. The library could be changed to add
 * flexibility, but I do not need this now, so I shall not bother.
 *
 * C99 states that, to be fully compliant, the only variables of static storage
 * duration which a signal handler may write to are those of type volatile
 * sig_atomic_t; furthermore, a signal handler is not allowed to read from any
 * such variable. Note, however, that C99 does not specify the concept of
 * signal masks, which are defined by POSIX. If you see the signal handler
 * below, you note that I actually read such a variable; I do not think this
 * constitutes a problem due to the fact that this handler is installed as
 * having its signal mask full, and the only place where this variable is
 * written to is in this same handler.
 */

#include <stddef.h>
#include <stdint.h>
#include <signal.h>
#include <unistd.h>
#include <fcntl.h>
#include <poll.h>

#if INT_MAX <= SIG_ATOMIC_MAX && INT_MIN >= SIG_ATOMIC_MIN
#define INT_FITS_SIG_ATOMIC 1
#else
#define INT_FITS_SIG_ATOMIC 0
#endif

volatile sig_atomic_t sig_caught = 0;
volatile static sig_atomic_t sig_caught_num;
#if INT_FITS_SIG_ATOMIC == 0
volatile static sig_atomic_t sig_caught_num_overflow = 0;
#endif
int sig_fd = -1;
static int sig_pipe[2];

static void sig_hand(int signum)
{
	/* This is a violation of C99 as we are not supposed to read anything;
	 * however, there should be no problem here, I believe, since the only
	 * place where this variable is written to is in this function (note it
	 * is declared as const in the sig.h) and we fill the signal mask for
	 * this signal when we install this handler later. Outside of a signal
	 * handler, this variable could be written to, but first all signals
	 * should be blocked and then unblocked after the write. */
	if (sig_caught)
		return;

#if INT_FITS_SIG_ATOMIC
	sig_caught = 1;
	sig_caught_num = (sig_atomic_t)signum;
#else
	if (signum <= SIG_ATOMIC_MAX && signum >= SIG_ATOMIC_MIN) {
		sig_caught = 1;
		sig_caught_num = (sig_atomic_t)signum;
	} else {
		sig_caught = 1;
		sig_caught_num_overflow = 1;
	}
#endif
	const unsigned char zero = '\0';
	while (write(sig_pipe[1], &zero, 1) == 0)
		;
}

/* Set default action and empty signal mask for array of signals. */
int sig_dfl_v(const int sigv[], size_t sigc)
{
	size_t i;
	struct sigaction sa = {0};
	sa.sa_handler = SIG_DFL;
	sigemptyset(&sa.sa_mask);
	for (i = 0; i < sigc; i++) {
		if (sigaction(sigv[i], &sa, NULL))
			return 1;
	}

	return 0;
}

/* Set default action and empty signal mask for signum. */
int sig_dfl(int signum)
{
	return sig_dfl_v(&signum, 1);
}

/* Make array of signals be ignored. */
int sig_ign_v(const int sigv[], size_t sigc)
{
	size_t i;
	struct sigaction sa = {0};
	sa.sa_handler = SIG_IGN;
	sigemptyset(&sa.sa_mask);
	for (i = 0; i < sigc; i++) {
		if (sigaction(sigv[i], &sa, NULL))
			return 1;
	}

	return 0;
}

/* Make signum be ignored. */
int sig_ign(int signum)
{
	return sig_ign_v(&signum, 1);
}

/* Add array of signals to the signals which are to be caught via the mechanism
 * described above. */
int sig_catch_v(const int sigv[], size_t sigc)
{ 
	size_t i;
	struct sigaction sa = {0};
	if (sig_fd == -1) {
		if (pipe(sig_pipe))
			return 1;
		sig_fd = sig_pipe[0];
		int flags = fcntl(sig_pipe[1], F_GETFL);
		if (flags == -1)
			return 1;
		if (fcntl(sig_pipe[1], F_SETFL, flags | O_NONBLOCK) == -1)
			return 1;
	}
	sa.sa_handler = sig_hand;
	/* Important that all signals which use this handler be added to the
	 * mask. Note that I read and write to the variable sig_caught in the
	 * handler. */
	sigfillset(&sa.sa_mask);
	for (i = 0; i < sigc; i++) {
		if (sigaction(sigv[i], &sa, NULL))
			return 1;
				
	}

	return 0;
}

/* Add signum to the signals which are to be caught via the mechanism described
 * above. */
int sig_catch(int signum)
{
	return sig_catch_v(&signum, 1);
}

/* Make signal set from array of signals. */
int sig_set_v(sigset_t *set, const int sigv[], size_t sigc)
{
	size_t i;
	sigemptyset(set);
	for (i = 0; i < sigc; i++) {
		if (sigaddset(set, sigv[i]))
			return 1;
	}
	
	return 0;
}

/* Remove array of signals from the signal mask of the process. */
int sig_unblock_v(const int sigv[], size_t sigc)
{
	sigset_t set;
	if (sig_set_v(&set, sigv, sigc))
		return 1;
	if (sigprocmask(SIG_UNBLOCK, &set, NULL))
		return 1;

	return 0;
}

/* Remove signum from the signal mask of the process. */
int sig_unblock(int signum)
{
	return sig_unblock_v(&signum, 1);
}

/* Add array of signals to the signal mask of the process. */
int sig_block_v(const int sigv[], size_t sigc)
{
	sigset_t set;
	if (sig_set_v(&set, sigv, sigc))
		return 1;
	if (sigprocmask(SIG_BLOCK, &set, NULL))
		return 1;

	return 0;
}

/* Add signum to the signal mask of the process. */
int sig_block(int signum)
{
	return sig_block_v(&signum, 1);
}

/* Add all signals to the signal mask of the process. */
int sig_block_all()
{
	sigset_t set;
	sigfillset(&set);
	return sigprocmask(SIG_BLOCK, &set, NULL) ? 1 : 0;
}

/* Reset the default action of the signal which was caught via the mechanism
 * described above and raise it. This is usually intended to terminate the
 * program. */
int sig_caught_raise()
{
	if (sig_caught == 0)
		return 2;
#if INT_FITS_SIG_ATOMIC == 0
	if (sig_caught_num_overflow)
		return 3;
#endif
	if (sig_dfl(sig_caught_num))
		return 1;
	if (raise(sig_caught_num))
		return 1;

	return 0;
}

/* Return the sinal number of the signal caught by the mechanism described
 * above. */
int sig_caught_num_get(int *signum)
{
	if (sig_caught == 0)
		return 2;
#if INT_FITS_SIG_ATOMIC == 0
	if (sig_caught_num_overflow)
		return 3;
#endif
	*signum = (int)sig_caught_num;

	return 0;
}

/* This function is to be used as a replacement for read. */
ssize_t sig_read(int fd, void *buf, size_t count)
{
	struct pollfd fds[2] = {
		{ .fd = sig_fd, .events = POLLIN},
		{ .fd = fd, .events = POLLIN}
	};
	if (poll(fds, sizeof fds / sizeof fds[0], -1) == -1)
		return -1;
	else if (fds[1].revents)
		return read(fd, buf, count);
	else
		return -1;
}

/* Returns 1 if any signal in the array sigv of length sigc is pending, 0 if
 * none is pending and -1 on error. */
int sig_pend_v(const int sigv[], size_t sigc)
{
	int res;
	size_t i;
	sigset_t set;
	if (sigpending(&set))
		return -1;
	for (i = 0; i < sigc; i++) {
		if ((res = sigismember(&set, sigv[i])))
			return res;
	}

	return 0;
}

/* Returns 1 if sig is pending 0 if it is not pending and -1 on error. */
int sig_pend(int sig)
{
	return sig_pend_v(&sig, 1);
}
