/*
 * Copyright (C) 2021 B.R.S. Roso.
 * 
 * Xselection is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License version 3 as published by the Free
 * Software Foundation.
 * 
 * This program is distributed in the hope that it be useful but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * Xselection.  If not, see <https://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <limits.h>
#include <poll.h>
#include <unistd.h>
#include <errno.h>
#include <stdint.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>

#include "perrf.h"
#include "lmem.h"
#include "sig.h"

const int sigv[] = { SIGHUP, SIGINT, SIGQUIT, SIGSEGV, SIGALRM, SIGTERM };
const size_t sigc = sizeof sigv / sizeof sigv[0];
unsigned int timeout = 60;
const char *selection = "PRIMARY";
const char *target = "UTF8_STRING";
const char *display_name;
struct lmem buf;
size_t buf_len;
int cont;
Display *d;
Window w;
Time t;
Atom target_atom;
Atom targets_atom;

/* We handle SIGALRM because the program needs to set an alarm to cause
 * termination, and we handle SIGTERM because I need other programs to be able
 * to tell this program to terminate. They need to be handled this way because
 * we want to erase the buffer before exiting. */
int sig_init()
{
	if (sig_block_all())
		return 1;
	if (sig_catch_v(sigv, sigc))
		return 1;
	if (sig_unblock_v(sigv, sigc))
		return 1;

	return 0;
}

void usage()
{
	perrf("usage: xselection [-c] [-t target] [-T seconds] [selection]");
}

int proc_args(int argc, char *argv[])
{
	int opt;
	unsigned long num;
	char *p;	
	while ((opt = getopt(argc, argv, "cd:t:T:")) != -1) {
		switch (opt) {
			case 'c':
				cont = 1;
				break;
			case 'd':
				display_name = optarg;
				break;
			case 't':
				target = optarg;
				break;
			case 'T':
				num = strtoul(optarg, &p, 10);
				if (*p) {
					usage();
					return 1;
				}
				if (num > UINT_MAX) {
					perrf("given timeout is too large");
					return 1;
				}
				timeout = num;
				break;
			default:
				usage();
				return 1;
		}
	}
	if (optind == argc - 1)
		selection = argv[optind];
	else if (optind != argc) {
		usage();
		return 1;
	}

	return 0;
}

/* Read all the data available from fd into locked memory. The data is returned
 * via buf_ret and its length via len_ret. This function allocates the buffer;
 * the caller MUST NOT have already allocated the buffer. In the event of an
 * error, buf_ret and len_ret are left untouched. */
int read_all(int fd, struct lmem *buf_ret, size_t *len_ret)
{
	int res = 0;
	ssize_t n;
	struct lmem buf = {0};
	size_t len = 0;

	if (lnew(&buf)) {
		res = 1;
		goto end;
	}
	while ((n = sig_read(fd, (char *)buf.mem + len, buf.size - len))) {
		if (sig_caught) {
			res = 1;
			goto end;
		}
		if (n < 0) {
			res = 1;
			goto end;
		}
		len += n;
		if (len == buf.size) {
			if (lgrow(&buf)) {
				res = 1;
				goto end;
			}
		}
	}

end:
	if (res) {
		if (ldestroy(&buf))
			perrf("WARNING: failed to correctly destroy mlock'd buffer");
	} else {
		*buf_ret = buf;
		*len_ret = len;
	}

	return res;
}

void deinit()
{
	if (ldestroy(&buf))
		perrf("WARNING: failed to correctly destroy mlock'd buffer");
	if (d)
		XCloseDisplay(d);
	if (sig_caught)
		sig_caught_raise();
}

int x_error_hand(Display *d, XErrorEvent *e)
{
	perrf("X error");
	deinit();
	exit(1);
}

int x_io_error_hand(Display *d)
{
	perrf("X error");
	deinit();
	exit(1);
}

int x_sel_req_event_hand(XSelectionRequestEvent *e)
{
	XEvent e2 = {0};
	int res;
	e2.xselection.type = SelectionNotify;
	e2.xselection.selection = e->selection;
	e2.xselection.target = e->target;
	e2.xselection.time = e->time;
	e2.xselection.requestor = e->requestor;
	const int valid_time = e->time >= t || e->time == CurrentTime;
	if (valid_time && e->target == target_atom) {
		e2.xselection.property = e->property;
		XChangeProperty(d, e->requestor, e->property, e->target,
				8, PropModeReplace, buf.mem, buf_len);
		res = !cont;
		perrf("selection sent to window %#llx",
				(unsigned long long)e->requestor);
	} else if (valid_time && e->target == targets_atom) {
		e2.xselection.property = e->property;
		Atom a[] = { targets_atom, target_atom };
		XChangeProperty(d, e->requestor, e->property, XA_ATOM,
				32, PropModeReplace,
				(unsigned char *)a, sizeof a / sizeof a[0]);
		res = 0;
		perrf("selection target list sent to window %#llx",
				(unsigned long long)e->requestor);
	} else {
		e2.xselection.property = None;
		res = 0;
		char * const name = XGetAtomName(d, e->target);
		perrf("incorrect target \"%s\" requested by window %#llx",
				name, (unsigned long long)e->requestor);
		XFree(name);
	}
	XSendEvent(d, e->requestor, False, NoEventMask, &e2);
	XFlush(d);
	return res;
}

int selection_init(Time timestamp)
{
	static int inited;
	if (inited)
		return 0;
	t = timestamp;
	const Atom s = XInternAtom(d, selection, False);
	XSetSelectionOwner(d, s, w, timestamp);
	if (w != XGetSelectionOwner(d, s)) {
		perrf("failed to become selection owner");
		return 1;
	}
	inited = 1;

	return 0;
}

int x_event_hand(XEvent *e)
{
	switch (e->type) {
		case PropertyNotify:
			selection_init(e->xproperty.time);
			return 0;
		case SelectionRequest:
			return x_sel_req_event_hand(&e->xselectionrequest);
		case SelectionClear:
			return 1;
		default:
			perrf("unknown X event received");
			return 0;
	}
}

void x_init()
{
	XSetWindowAttributes a = { .event_mask = PropertyChangeMask };
	XSetErrorHandler(x_error_hand);
	XSetIOErrorHandler(x_io_error_hand);
	d = XOpenDisplay(display_name);
	const Window r = DefaultRootWindow(d);
	w = XCreateWindow(d, r, 0, 0, 1, 1, 0, CopyFromParent, InputOnly,
			CopyFromParent, CWEventMask, &a);
	const Atom init_atom = XInternAtom(d, "INIT", False);
	const Atom type_atom = XInternAtom(d, "STRING", False);
	target_atom = XInternAtom(d, target, False);
	targets_atom = XInternAtom(d, "TARGETS", False);
	XChangeProperty(d, w, init_atom, type_atom, 8, PropModeAppend, NULL, 0);
	XFlush(d);
}

int loop()
{
	XEvent e;
	struct pollfd fds[2];
	fds[0].fd = XConnectionNumber(d);
	fds[0].events = POLLIN;
	fds[1].fd = sig_fd;
	fds[1].events = POLLIN;
	do {
		if (XPending(d) > 0) {
			XNextEvent(d, &e);
			if (x_event_hand(&e))
				return 0;

		}
		if (poll(fds, sizeof fds / sizeof fds[0], -1) == -1) {
			if (errno != EINTR)
				return 1;
		}
		if (sig_caught)
			return 0;
	} while (1);

	return 0;
}

int main(int argc, char *argv[])
{
	int res;
	perrf_head = argv[0];
	if ((res = proc_args(argc, argv)))
		goto end;
	if ((res = sig_init()))
		goto end;
	if ((res = read_all(0, &buf, &buf_len))) {
		if (sig_caught)
			perrf("deadly signal received while reading input");
		else
			perrf("error reading input");
		goto end;
	}
	alarm(timeout);
	x_init();
	res = loop();

end:
	deinit();
	return res;
}
